"use strict";

exports.__esModule = true;
exports.getOrientation = getOrientation;
exports.convertGif = convertGif;
exports.takePic = takePic;
exports.transformPic = transformPic;
exports.mediaDevices = exports.webcamSupport = exports.webPSupport = void 0;

var _isNode = _interopRequireDefault(require("./is-node"));

var _pica = _interopRequireDefault(require("./pica"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var canvas, ctx, _webPSupport, _webcamSupport, _mediaDevices;

if (!_isNode.default) {
  if (!HTMLCanvasElement.prototype.toBlob) {
    HTMLCanvasElement.prototype.toBlob = function (callback, type, quality) {
      var dataURL = this.toDataURL(type, quality).split(',')[1];
      setTimeout(function () {
        var binStr = atob(dataURL);
        var len = binStr.length;
        var arr = new Uint8Array(len);

        for (var i = 0; i < len; ++i) {
          arr[i] = binStr.charCodeAt(i);
        }

        callback(new Blob([arr], {
          type: type || 'image/png'
        }));
      });
    };
  }

  window.URL = window.URL || window.webkitURL || window.mozURL || window.msURL;
  canvas = document.createElement('canvas');
  ctx = canvas.getContext('2d');
  canvas.width = canvas.height = 1;
  _webPSupport = 5 === canvas.toDataURL('image/webp').indexOf('image/webp');
  _webcamSupport = navigator.mediaDevices && navigator.mediaDevices.getUserMedia;

  _mediaDevices = function () {
    var md = navigator.mediaDevices || navigator;
    md.getMedia = md.getUserMedia || md.webkitGetUserMedia || md.mozGetUserMedia || md.msGetUserMedia;
    return md;
  }();
}

var webPSupport = _webPSupport;
exports.webPSupport = webPSupport;
var webcamSupport = _webcamSupport;
exports.webcamSupport = webcamSupport;
var mediaDevices = _mediaDevices; // Get orientation of a given pic
// https://stackoverflow.com/questions/7584794/accessing-jpeg-exif-rotation-data-in-javascript-on-the-client-side/32490603#32490603

exports.mediaDevices = mediaDevices;
var iOs = !_isNode.default && /iPad|iPhone|iPod/.test(navigator.platform);

function getOrientation(file, callback) {
  if (iOs) return callback(0); // iOs always display images with correct orientation, no need to do it manually

  var reader = new FileReader();

  reader.onerror = function (event) {
    reader.abort();
    callback(-1);
  };

  reader.onload = function (e) {
    var view = new DataView(e.target.result);
    if (view.getUint16(0, false) != 0xffd8) return callback(-2);
    var length = view.byteLength;
    var offset = 2;
    var marker, little, tags, i;

    while (offset < length) {
      if (view.getUint16(offset + 2, false) <= 8) return callback(-1);
      marker = view.getUint16(offset, false);
      offset += 2;

      if (marker == 0xffe1) {
        if (view.getUint32(offset += 2, false) != 0x45786966) return callback(-1);
        little = view.getUint16(offset += 6, false) == 0x4949;
        offset += view.getUint32(offset + 4, little);
        tags = view.getUint16(offset, little);
        offset += 2;

        for (i = 0; i < tags; ++i) {
          if (view.getUint16(offset + i * 12, little) == 0x0112) return callback(view.getUint16(offset + i * 12 + 8, little));
        }
      } else if ((marker & 0xff00) != 0xff00) {
        break;
      } else {
        offset += view.getUint16(offset, false);
      }
    }

    return callback(-1);
  };

  reader.readAsArrayBuffer(file.slice(0, 64 * 4 * 1024));
}

function convertGif(file, callback, error) {
  var img = new Image();
  img.onerror = error;

  img.onload = function () {
    URL.revokeObjectURL(file);
    canvas.width = img.width;
    canvas.height = img.height;
    ctx.drawImage(img, 0, 0);
    callback(canvas.toDataURL());
  };

  img.src = URL.createObjectURL(file);
} // Take a snapshot of the video stream


function takePic(video, callback) {
  canvas.width = video.videoWidth;
  canvas.height = video.videoHeight;
  ctx.drawImage(video, 0, 0);
  canvas.toBlob(callback, 'image/png');
}

function transformPic(image, options, callback) {
  canvas.width = options.width;
  canvas.height = options.height;
  var nw2 = image.naturalWidth / 2;
  var nh2 = image.naturalHeight / 2;
  var rad = options.rotate * Math.PI / 180;
  var x1 = -nw2 + options.left;
  var y1 = -nh2 + options.top;
  var x2 = -nw2 + (options.left + options.width);
  var y2 = -nh2 + (options.top + options.height);
  var coord1 = newCoord(x1, y1, rad);
  var coord2 = newCoord(x2, y2, rad);
  var left = Math.min(coord1.x, coord2.x) + nw2;
  var top = Math.min(coord1.y, coord2.y) + nh2;
  var right = Math.max(coord1.x, coord2.x) + nw2;
  var bottom = Math.max(coord1.y, coord2.y) + nh2;
  var width = right - left;
  var height = bottom - top;
  var add = newSize(width, height, rad);
  var addWidth = Math.abs(add.width - width);
  var addHeight = Math.abs(add.height - height);
  left -= addWidth;
  top -= addHeight;
  width += addWidth * 2;
  height += addHeight * 2;
  if (options.flip.h < 0) left = image.naturalWidth - left - width;
  if (options.flip.v < 0) top = image.naturalHeight - top - height;
  ctx.save();
  ctx.translate(canvas.width / 2, canvas.height / 2);
  ctx.rotate(Math.PI / 180 * options.rotate);
  ctx.scale(options.flip.h, options.flip.v);
  ctx.drawImage(image, left, top, width, height, -width / 2, -height / 2, width, height);
  ctx.restore();

  var downsizeCallback = function downsizeCallback() {
    if (options.webp && _webPSupport) {
      canvas.toBlob(callback, 'image/webp');
    } else {
      canvas.toBlob(callback, 'image/jpeg', 0.92);
    }
  };

  if (options.resizeToWidth) {
    (0, _pica.default)(canvas, ctx, options.resizeToWidth, options.resizeToHeight, downsizeCallback);
  } else {
    downsizeCallback();
  }
}

function newCoord(x, y, rad) {
  var sin = Math.sin(rad);
  var cos = Math.cos(rad);
  return {
    x: x * cos + y * sin,
    y: -x * sin + y * cos
  };
}

function newSize(width, height, rad) {
  var cos = Math.abs(Math.cos(rad));
  var sin = Math.abs(Math.sin(rad));
  return {
    width: height * sin + width * cos,
    height: height * cos + width * sin
  };
}